/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

namespace fe
{

FlatTree::FlatTree(void):
	m_pVertexArray(NULL),
	m_vertices(0),
	m_pSphereArray(NULL),
	m_grain(3)
{
//	feLog("FlatTree::FlatTree\n");
}

void FlatTree::populate(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,U32 a_vertices,
	const SpatialVector& a_center,F32 a_radius)
{
	//* TODO cache a copy?
	m_pVertexArray=a_pVertexArray;
	m_pNormalArray=a_pNormalArray;
	m_vertices=a_vertices;

	cacheSpheres();
}

void FlatTree::cacheSpheres(void)
{
	U32 sphereCount=m_vertices/m_grain;
	m_pSphereArray=(Vector<4,Real>*)reallocate(m_pSphereArray,
			sphereCount*sizeof(Vector<4,Real>));
	for(U32 m=0;m<sphereCount;m++)
	{
		SpatialVector center;
		Real radius;
		BoundingSphere<Real>::solve(&m_pVertexArray[m*3],3,center,radius);
		m_pSphereArray[m]=center;
		m_pSphereArray[m][3]=radius;

//		for(U32 n=0;n<3;n++)
//		{
//			feLog("  %s\n",c_print(m_pVertexArray[m*3+n]));
//		}
//		feLog("SurfaceTriangles::cacheSpheres %d %s\n",
//				m,c_print(m_pSphereArray[m]));
	}
}

Real FlatTree::nearestPoint(const SpatialVector& a_origin,U32 a_hitLimit,
	std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray) const
{
	for(U32 m=0;m<m_vertices;m+=m_grain)
	{
		const U32 sphereIndex=m/m_grain;
		nearestAccumulate(a_origin,a_hitLimit,a_rHitArray,
				m_pSphereArray[sphereIndex],m_pSphereArray[sphereIndex][3],
				m_pVertexArray[m],m_pVertexArray[m+1],m_pVertexArray[m+2],
				m_pNormalArray[m],m_pNormalArray[m+1],m_pNormalArray[m+2]);
	}
	return a_rHitArray.size()? a_rHitArray[0]->distance(): 0.0;
}

void FlatTree::nearestAccumulate(const SpatialVector& a_origin,U32 a_hitLimit,
	std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray,
	const SpatialVector& a_rCenter,Real a_radius,
	const SpatialVector& a_rVertex0,const SpatialVector& a_rVertex1,
	const SpatialVector& a_rVertex2,const SpatialVector& a_rNormal0,
	const SpatialVector& a_rNormal1,const SpatialVector& a_rNormal2) const
{
	U32 hits=a_rHitArray.size();
	Real margin_distance=hits? a_rHitArray[hits-1]->distance(): -1.0;

	SpatialVector direction;
	if(margin_distance>=0.0)
	{
		const Real distance=PointSphereNearest<Real>::solve(a_rCenter,a_radius,
				a_origin,direction);
		if(distance>margin_distance)
		{
			return;
		}
	}

	Barycenter<Real> barycenter;
	SpatialVector intersection;
	const Real distance=
			PointTriangleNearest<Real>::solve(a_rVertex0,a_rVertex1,a_rVertex2,
			a_origin,direction,intersection,barycenter);
	if(distance>=0 && (margin_distance<0 || distance<margin_distance))
	{
		addHit(a_hitLimit,a_rHitArray,m_hitPool,
				distance,direction,intersection,barycenter,
				a_rVertex0,a_rVertex1,a_rVertex2,
				a_rNormal0,a_rNormal1,a_rNormal2);
	}
}

void FlatTree::addHit(U32 a_hitLimit,
	std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray,
	CountedPool<Hit>& a_rHitPool,Real a_distance,
	const SpatialVector& a_rDirection,
	const SpatialVector& a_rIntersection,
	const Barycenter<Real>& a_rBarycenter,
	const SpatialVector& a_rVertex0,const SpatialVector& a_rVertex1,
	const SpatialVector& a_rVertex2,const SpatialVector& a_rNormal0,
	const SpatialVector& a_rNormal1,const SpatialVector& a_rNormal2)
{
	U32 hits=a_rHitArray.size();
	if(hits<a_hitLimit)
	{
		hits++;
		a_rHitArray.resize(hits);
		a_rHitArray[hits-1]=a_rHitPool.get();
	}
	sp<Hit> spHit(a_rHitArray[hits-1]);
	spHit->m_pVertex[0]=a_rVertex0;
	spHit->m_pVertex[1]=a_rVertex1;
	spHit->m_pVertex[2]=a_rVertex2;
	spHit->m_pNormal[0]=a_rNormal0;
	spHit->m_pNormal[1]=a_rNormal1;
	spHit->m_pNormal[2]=a_rNormal2;
	spHit->m_barycenter=a_rBarycenter;
	spHit->m_intersection=a_rIntersection;
	spHit->m_direction=a_rDirection;
	spHit->m_distance=a_distance;

	std::sort(a_rHitArray.begin(),a_rHitArray.end(),Sort());
}

Real FlatTree::rayImpact(const SpatialVector& a_origin,
	const SpatialVector& a_direction,U32 a_hitLimit,
	std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray) const
{
	Real best_distance= -1.0f;
	for(U32 m=0;m<m_vertices;m+=m_grain)
	{
		const SpatialVector center=m_pSphereArray[m/m_grain];
		const Real radius=m_pSphereArray[m/m_grain][3];
		Real distance=RaySphereIntersect<Real>::solve(center,radius,
				a_origin,a_direction);
		if(distance<0.0f)
		{
			continue;
		}

		Barycenter<Real> barycenter;
		distance=RayTriangleIntersect<Real>::solve(
				m_pVertexArray[m],
				m_pVertexArray[m+1],
				m_pVertexArray[m+2],
				a_origin,a_direction,barycenter);
		if(distance>=0.0 && (best_distance<0.0f ||
				distance<best_distance))
		{
			best_distance=distance;

			sp<Hit> spHit=m_hitPool.get();
			a_rHitArray.resize(1);
			a_rHitArray[0]=spHit;
			spHit->m_pVertex[0]=m_pVertexArray[m];
			spHit->m_pVertex[1]=m_pVertexArray[m+1];
			spHit->m_pVertex[2]=m_pVertexArray[m+2];
			spHit->m_pNormal[0]=m_pNormalArray[m];
			spHit->m_pNormal[1]=m_pNormalArray[m+1];
			spHit->m_pNormal[2]=m_pNormalArray[m+2];
			spHit->m_barycenter=barycenter;
			spHit->m_direction=a_direction;
			spHit->m_intersection=a_origin+distance*a_direction;
			spHit->m_distance=distance;
		}
	}

	return best_distance;
}

} // namespace
