/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_ST_DEBUG				FALSE

#define	FE_ST_TREE_DRAW_DEBUG	TRUE
#define	FE_ST_TREE_DRAW_NORMALS	FALSE

#define	FE_ST_TREE_HIT_DEBUG	FALSE

namespace fe
{

SurfaceTriangles::SurfaceTriangles(void):
	m_vertices(0),
	m_pVertexArray(NULL),
	m_pNormalArray(NULL)
{
	m_trianglePower.setIterations(8);
}

SurfaceTriangles::~SurfaceTriangles(void)
{
	deallocate(m_pNormalArray);
	deallocate(m_pVertexArray);
}

Protectable* SurfaceTriangles::clone(Protectable* pInstance)
{
#if	FE_ST_DEBUG
	feLog("SurfaceTriangles::clone\n");
#endif

	SurfaceTriangles* pSurfaceTriangles= pInstance?
			dynamic_cast<SurfaceTriangles*>(pInstance): new SurfaceTriangles();

	checkCache();

	SurfaceSphere::clone(pSurfaceTriangles);

	// TODO copy attributes
	const U32 size=m_vertices*sizeof(SpatialVector);

	pSurfaceTriangles->m_vertices=m_vertices;

	pSurfaceTriangles->m_pVertexArray=(SpatialVector*)reallocate(
			pSurfaceTriangles->m_pVertexArray,size);
	memcpy(pSurfaceTriangles->m_pVertexArray,m_pVertexArray,size);

	pSurfaceTriangles->m_pNormalArray=(SpatialVector*)reallocate(
			pSurfaceTriangles->m_pNormalArray,size);
	memcpy(pSurfaceTriangles->m_pNormalArray,m_pNormalArray,size);

	return pSurfaceTriangles;
}

void SurfaceTriangles::cache(void)
{
	feLog("SurfaceTriangles::cache\n");

	if(!m_record.isValid())
	{
		feLog("SurfaceTriangles::cache record invalid\n");
		return;
	}

	SurfaceSphere::cache();

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(m_record);

	sp<RecordGroup> spRG=surfaceModelRV.surfacePointRG();
	for(RecordGroup::iterator it=spRG->begin();it!=spRG->end();it++)
	{
		sp<RecordArray> spRA= *it;

		SurfacePoint m_surfacePointRV;
		m_surfacePointRV.bind(spRA);
		while(m_surfacePointRV.step())
		{
			SpatialVector& location=m_surfacePointRV.location();
			SpatialVector& up=m_surfacePointRV.up();
//			SpatialVector& texCoord=m_surfacePointRV.texCoord();

			m_pVertexArray=(SpatialVector*)reallocate(m_pVertexArray,
					(m_vertices+1)*sizeof(SpatialVector));
			m_pNormalArray=(SpatialVector*)reallocate(m_pNormalArray,
					(m_vertices+1)*sizeof(SpatialVector));

			SpatialVector& rVertex=m_pVertexArray[m_vertices];
			SpatialVector& rNormal=m_pNormalArray[m_vertices];

			rVertex=location;
			rNormal=up;

			printf("vertex %d  %s  %s\n",m_vertices,
					c_print(rVertex),c_print(rNormal));
			m_vertices++;
		}
	}

	if(surfaceModelRV.searchComponent().isValid())
	{
		m_spSpatialTreeI=surfaceModelRV.searchComponent();
	}
}

void SurfaceTriangles::cacheResolve(void)
{
#if	FE_ST_DEBUG
	feLog("SurfaceTriangles::cacheResolve m_vertices=%d tree valid=%d\n",
			m_vertices,m_spSpatialTreeI.isValid());
#endif

	if(!m_vertices)
	{
		return;
	}
	if(!m_spSpatialTreeI.isValid())
	{
		m_spSpatialTreeI=create("SpatialTreeI.default");
	}
	if(!m_spSpatialTreeI.isValid())
	{
		return;
	}

#if	FE_ST_DEBUG
	feLog("SurfaceTriangles::cacheResolve building spatial tree\n");
#endif

	m_spSpatialTreeI->setRefinement(m_refinement);
	m_spSpatialTreeI->populate(m_pVertexArray,m_pNormalArray,
			m_vertices,center(),radius());
}

SpatialTransform SurfaceTriangles::sample(Vector2 a_uv) const
{
	SpatialTransform result;
	setIdentity(result);
	return result;
}

void SurfaceTriangles::drawInternal(BWORD a_transformed,
	const SpatialTransform* a_pTransform,sp<DrawI> a_spDrawI,
	const fe::Color* a_pColor) const
{
//	return;

#if FE_ST_TREE_DRAW_DEBUG
	sp<DrawMode> spWireframe(new DrawMode);
	spWireframe->copy(a_spDrawI->drawMode());
	spWireframe->setDrawStyle(DrawMode::e_outline);
	a_spDrawI->pushDrawMode(spWireframe);
#endif

#if TRUE
	if(m_vertices)
	{
		const Color white(1.0f,1.0f,1.0f,1.0f);

		const Vector2f* pTexture=NULL;
		const BWORD multicolor=FALSE;
		if(a_transformed)
		{
			a_spDrawI->drawTransformedTriangles(*a_pTransform,NULL,
					m_pVertexArray,m_pNormalArray,pTexture,m_vertices,
					DrawI::e_discrete,multicolor,a_pColor? a_pColor: &white);
		}
		else
		{
			a_spDrawI->drawTriangles(m_pVertexArray,m_pNormalArray,pTexture,
					m_vertices,DrawI::e_discrete,multicolor,
					a_pColor? a_pColor: &white);
		}
	}
	else
	{
		feLog("SurfaceTriangles::draw no vertices\n");
	}
#endif

#if FE_ST_TREE_DRAW_NORMALS
	if(m_vertices)
	{
		const Color red(1.0f,0.0f,0.0f,1.0f);
		const Real normalScale=1.0;
		SpatialVector pLines[2];
		for(U32 m=0;m<m_vertices;m++)
		{
			pLines[0]=m_pVertexArray[m];
			pLines[1]=pLines[0]+m_pNormalArray[m]*normalScale;

			const SpatialVector* pNormals=NULL;
			const BWORD multicolor=FALSE;
			if(a_transformed)
			{
				a_spDrawI->drawTransformedLines(*a_pTransform,NULL,pLines,
						pNormals,2,DrawI::e_discrete,multicolor,&red);
			}
			else
			{
				a_spDrawI->drawLines(pLines,pNormals,2,DrawI::e_discrete,
						multicolor,&red);
			}
		}
	}
#endif

#if FE_ST_TREE_DRAW_DEBUG
	a_spDrawI->popDrawMode();
#endif

#if FE_ST_TREE_DRAW_DEBUG
	if(m_spSpatialTreeI.isValid())
	{
		m_spSpatialTreeI->draw(a_spDrawI,NULL);
	}
#endif
}

sp<SurfaceTriangles::ImpactI> SurfaceTriangles::nearestPoint(
		const SpatialVector& a_origin) const
{
	if(!m_vertices || !m_spSpatialTreeI.isValid())
	{
		return sp<ImpactI>(NULL);
	}

	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const U32 hitLimit=32;
	m_spSpatialTreeI->nearestPoint(a_origin,hitLimit,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount)
	{
		feLog("SurfaceTriangles::nearestPoint no hits\n");
		return sp<ImpactI>(NULL);
	}

#if FE_ST_TREE_HIT_DEBUG
	feLog("\n");
	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		feLog("Hit %d/%d dist %.6G  inter %s\n",
				m,hitCount,rspHit->distance(),c_print(rspHit->intersection()));
		for(U32 n=0;n<3;n++)
		{
			feLog("  v%d %s n %s\n",n,c_print(rspHit->vertex()[n]),
					c_print(rspHit->normal()[n]));
/*
			FEASSERT(fabs(magnitude(rspHit->vertex()[n]))<1e9);
			FEASSERT(fabs(magnitude(rspHit->normal()[n]))<1.1);
			if(fabs(magnitude(rspHit->vertex()[n]))>1e9 ||
					fabs(magnitude(rspHit->normal()[n]))>1.1)
			{
				exit(1);
			}
*/
		}
	}
#endif

	Real hitWeight[hitLimit];
	hitWeight[0]=1.0;

#if TRUE
	//* simply closest
	const Real sum=1.0;
	for(U32 m=1;m<hitCount;m++)
	{
		hitWeight[m]=0.0;
	}
#endif

#if FALSE
	//* blend by dot to normal
	const Real power=1.0;	// tweak
	Real sum=0.0;
	for(U32 m=0;m<hitCount;m++)
	{
		const sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		const SpatialVector to=unit(a_origin-rspHit->intersection());
		const SpatialVector norm=unit(location(rspHit->barycenter(),
				rspHit->normal()[0],rspHit->normal()[1],rspHit->normal()[2]));
		const Real to_dot=dot(to,norm);
		hitWeight[m]=pow(to_dot>0.0? to_dot: 0.0,power);
		sum+=hitWeight[m];
		feLog("Weight %d %.6G/%.6G to %s\n",m,hitWeight[m],sum,c_print(to));
	}
#endif

#if FALSE
	//* blend by distance
	const Real dist0=hitArray[0]->distance();
	const Real power=32.0;	// tweak
	Real sum=1.0;
	for(U32 m=1;m<hitCount;m++)
	{
		hitWeight[m]=pow(dist0/hitArray[m]->distance(),power);
		sum+=hitWeight[m];
//		feLog("Weight %d %.6G/%.6G\n",m,hitWeight[m],sum);
	}
#endif

	// TODO pick single best hit
	U32 bestHit=0;
	Real bestScore=0.0;
	for(U32 m=0;m<hitCount;m++)
	{
		const sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		Real thisScore=0.0;
		const SpatialVector diff=a_origin-rspHit->intersection();
		const Real mag=magnitude(diff);
		if(mag>1e-6)
		{
			const SpatialVector to=diff/mag;
			const SpatialVector norm=unitSafe(location(rspHit->barycenter(),
					rspHit->normal()[0],rspHit->normal()[1],
					rspHit->normal()[2]));
			const Real to_dot=dot(to,norm);
			thisScore=(to_dot>0.0)? to_dot: 0.0;
#if FE_ST_TREE_HIT_DEBUG
			feLog("  to %s\n",c_print(to));
#endif
		}
		if(!m)
		{
			bestScore=thisScore;
		}
		else if(thisScore>bestScore)
		{
			bestScore=thisScore;
			bestHit=m;
		}
#if FE_ST_TREE_HIT_DEBUG
		feLog("Score %d/%d %.6G best %.6G\n",
				m,hitCount,thisScore,bestScore);
#endif
	}

	sp<SpatialTreeI::Hit>& rspHit=hitArray[bestHit];
	sp<Impact> spImpact=m_triangleImpactPool.get();
	spImpact->setSurface(dynamic_cast<const SurfaceTriangles*>(this));
	spImpact->setOrigin(a_origin);
	spImpact->setDirection(rspHit->direction());
	spImpact->setDistance(rspHit->distance());
	spImpact->setBarycenter(rspHit->barycenter());
	spImpact->setVertex0(rspHit->vertex()[0]);
	spImpact->setVertex1(rspHit->vertex()[1]);
	spImpact->setVertex2(rspHit->vertex()[2]);
	spImpact->setNormal0(rspHit->normal()[0]);
	spImpact->setNormal1(rspHit->normal()[1]);
	spImpact->setNormal2(rspHit->normal()[2]);

	//* reevaluate barycenter based on normals
//	const SpatialVector to=unit(a_origin-rspHit->intersection());
	const SpatialVector norm=unitSafe(location(rspHit->barycenter(),
			rspHit->normal()[0],rspHit->normal()[1],rspHit->normal()[2]));
	Barycenter<Real> barynormal;
#if FALSE
	const Real dot0=dot(norm,rspHit->normal()[0]);
	const Real dot1=dot(norm,rspHit->normal()[1]);
	const Real dot2=dot(norm,rspHit->normal()[2]);
	const SpatialVector zero(0.0f,0.0f,0.0f);
	const SpatialVector part0=rspHit->normal()[0]-dot0*norm;
	const SpatialVector part1=rspHit->normal()[1]-dot1*norm;
	const SpatialVector part2=rspHit->normal()[2]-dot2*norm;
	barynormal.solve(part0,part1,part2,zero);
#else
//	feLog("n0 %s\n",c_print(rspHit->normal()[0]));
//	feLog("n1 %s\n",c_print(rspHit->normal()[1]));
//	feLog("n2 %s\n",c_print(rspHit->normal()[2]));
//	feLog("norm %s\n",c_print(norm));
	barynormal.solve(rspHit->normal()[0],rspHit->normal()[1],
			rspHit->normal()[2],norm);
//	feLog("barynormal %s\n",c_print(barynormal));
#endif
	spImpact->setBarycenter(barynormal);
	const SpatialVector intersection=location(barynormal,
			rspHit->vertex()[0],rspHit->vertex()[1],rspHit->vertex()[2]);
	const SpatialVector toward=intersection-a_origin;
	const Real range=magnitude(toward);
	spImpact->setDistance(range);
	spImpact->setDirection(range>0.0? toward/range: toward);

	if(hitCount>1)
	{
		SpatialVector intersection;
		SpatialVector normal;
		set(intersection);
		set(normal);
		for(U32 m=0;m<hitCount;m++)
		{
			const Real fraction=hitWeight[m]/sum;
			intersection+=fraction*hitArray[m]->intersection();

			const SpatialVector* norms=hitArray[m]->normal();
			const Vector2& bary=hitArray[m]->barycenter();
			normal+=fraction*(norms[0]*bary[0]+norms[1]*bary[1]+
					norms[2]*(1.0f-bary[0]-bary[1]));
		}
		normalizeSafe(normal);
		const SpatialVector to=intersection-a_origin;
		const Real mag=magnitude(to);
		spImpact->setDistance(mag);
		spImpact->setDirection(mag>1e-6? to/mag: to);

		// HACK horribly redundant, setup for impact resolution
		spImpact->setNormal0(normal);
		spImpact->setNormal1(normal);
		spImpact->setNormal2(normal);

#if FE_ST_TREE_HIT_DEBUG
		feLog("Blend %.6G %s\n",
				spImpact->distance(),c_print(intersection));
#endif
	}

	return spImpact;
}

sp<SurfaceTriangles::ImpactI> SurfaceTriangles::rayImpact(
		const SpatialVector& a_origin,const SpatialVector& a_direction) const
{
	if(!m_vertices || !m_spSpatialTreeI.isValid())
	{
		return sp<ImpactI>(NULL);
	}

	const U32 hitLimit=3;
	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	m_spSpatialTreeI->rayImpact(a_origin,a_direction,hitLimit,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount)
	{
		return sp<ImpactI>(NULL);
	}

	sp<SpatialTreeI::Hit>& rspHit=hitArray[0];

	sp<Impact> spImpact=m_triangleImpactPool.get();
	spImpact->setSurface(dynamic_cast<const SurfaceTriangles*>(this));
	spImpact->setOrigin(a_origin);
	spImpact->setDirection(a_direction);
	spImpact->setDistance(rspHit->distance());
	spImpact->setBarycenter(rspHit->barycenter());
	spImpact->setVertex0(rspHit->vertex()[0]);
	spImpact->setVertex1(rspHit->vertex()[1]);
	spImpact->setVertex2(rspHit->vertex()[2]);
	spImpact->setNormal0(rspHit->normal()[0]);
	spImpact->setNormal1(rspHit->normal()[1]);
	spImpact->setNormal2(rspHit->normal()[2]);
	return spImpact;
}

void SurfaceTriangles::resolveImpact(sp<ImpactI> a_spImpactI) const
{
	sp<Impact> spImpact(a_spImpactI);

	// TEMP if no nearest
	if(!spImpact.isValid() || spImpact->distance()<Real(0))
	{
		return;
	}

	SpatialVector intersection;
	SpatialVector normal;
#if FALSE
	m_trianglePower.solve(
			spImpact->vertex0(),
			spImpact->vertex1(),
			spImpact->vertex2(),
			spImpact->normal0(),
			spImpact->normal1(),
			spImpact->normal2(),
			spImpact->uv(),
			intersection,normal);
#else
	RayTriangleIntersect<Real>::resolveContact(
			spImpact->vertex0(),
			spImpact->vertex1(),
			spImpact->vertex2(),
			spImpact->normal0(),
			spImpact->normal1(),
			spImpact->normal2(),
			spImpact->origin(),
			spImpact->direction(),
			spImpact->barycenter(),
			spImpact->distance(),
			intersection,normal);
#endif
	spImpact->setIntersectionLocal(intersection);
	spImpact->setNormalLocal(normal);
}

} // namespace
